// Test application for the ThreadSafeProxy.
//
// Mike Woodring
// http://staff.develop.com/woodring
//
using System;
using System.Threading;

[ThreadSafeProxy]
class Accumulator : ContextBoundObject // MarshalByRefObject
{
    public Accumulator()
        : this(0)
    {
    }

    public Accumulator( int initValue )
    {
        sum = initValue;
        Console.WriteLine("Accumulator.ctor");
    }

    public void Add( int n )
    {
        int temp = sum;
        temp += n;
        Thread.Sleep(0); // An ill-timed loss of the processor...
        sum = temp;
    }

    public int Value
    {
        get { return(sum); }
    }

    private int sum = 0;
}

class App
{
    // Do this for MarshalByRefObject types:
    //   static Accumulator acc = (Accumulator)ThreadSafeProxy.Wrap(new Accumulator());

    // Do this for ContextBoundObject types that have
    // the ThreadSafeProxy attribute applied:
    //
    static Accumulator acc = new Accumulator();

    // To see the bug that results from the lack of thread synchronization,
    // use the above call to operator new, but remove the ThreadSafeProxy attribute.
    // The final value of the accumulator when the program terminates should always
    // be 5.  Without proper synchronization, it won't be.
    //
    static void Main()
    {
        Thread[] threads = new Thread[5];

        for( int n = 0; n < threads.Length; n++ )
        {
            threads[n] = new Thread(new ThreadStart(CallAdd));
            threads[n].Start();
        }

        for( int n = 0; n < threads.Length; n++ )
        {
            threads[n].Join();
        }

        Console.WriteLine("Accumulator: {0}", acc.Value);

        // Test ISynchronize.
        //
        ISynchronize s = acc as ISynchronize;
        
        if( s != null )
        {
            lock( s.SyncRoot )
            {
                Console.WriteLine("Accumulator: {0}", acc.Value);
            }
        }
    }

    static void CallAdd()
    {
        acc.Add(1);
    }
}
